use anyhow::{Context, Result};
use galactica_content::Content;
use galactica_system::phys::PhysSimShipHandle;
use glyphon::TextArea;
use log::{debug, error, trace};
use rhai::{Array, Dynamic, Engine, Scope};
use std::{collections::HashSet, num::NonZeroU32, rc::Rc};

use super::{
	api::{
		self, MouseClickEvent, MouseHoverEvent, PlayerShipStateEvent, SceneAction, SceneConfig,
		SpriteElement, TextBoxBuilder,
	},
	event::Event,
	util::{FpsIndicator, RadialBar, TextBox},
	UiElement,
};
use crate::{
	ui::{
		api::{RadialBuilder, SpriteBuilder, State},
		util::Sprite,
	},
	RenderInput, RenderState,
};

pub(crate) struct UiManager {
	current_scene: Option<String>,
	current_scene_config: SceneConfig,
	engine: Engine,
	scope: Scope<'static>,
	elements: Vec<UiElement>,
	ct: Rc<Content>,

	last_player_state: u32,
	show_timings: bool,
	fps_indicator: FpsIndicator,
}

impl UiManager {
	pub fn new(ct: Rc<Content>, state: &mut RenderState) -> Self {
		let scope = Scope::new();

		let mut engine = Engine::new();
		api::register_into_engine(&mut engine);

		Self {
			ct,
			current_scene: None,
			current_scene_config: SceneConfig::new(),
			engine,
			scope,
			elements: Vec::new(),
			show_timings: true,
			fps_indicator: FpsIndicator::new(state),
			last_player_state: 0,
		}
	}

	pub fn get_config(&self) -> &SceneConfig {
		&self.current_scene_config
	}

	/// Change the current scene
	pub fn set_scene(
		&mut self,
		state: &mut RenderState,
		input: Rc<RenderInput>,
		scene: String,
	) -> Result<()> {
		if !self.ct.get_config().ui_scenes.contains_key(&scene) {
			error!("tried to switch to ui scene `{scene}`, which doesn't exist");
			return Ok(());
		}

		debug!("switching to {:?}", scene);
		self.current_scene = Some(scene);
		self.current_scene_config = self
			.engine
			.call_fn(
				&mut self.scope,
				&self
					.ct
					.get_config()
					.ui_scenes
					.get(self.current_scene.as_ref().unwrap())
					.unwrap(),
				"config",
				(),
			)
			.with_context(|| format!("while handling `config()`"))
			.with_context(|| format!("in ui scene `{}`", self.current_scene.as_ref().unwrap()))?;

		self.scope.clear();
		self.elements.clear();
		let mut used_names = HashSet::new();

		let builders: Array = self
			.engine
			.call_fn(
				&mut self.scope,
				&self
					.ct
					.get_config()
					.ui_scenes
					.get(self.current_scene.as_ref().unwrap())
					.unwrap(),
				"init",
				(State::new(input.clone()),),
			)
			.with_context(|| format!("while running `init()`"))
			.with_context(|| format!("in ui scene `{}`", self.current_scene.as_ref().unwrap()))?;

		trace!("found {:?} builders", builders.len());

		for v in builders {
			if v.is::<SpriteBuilder>() {
				let s = v.cast::<SpriteBuilder>();
				if used_names.contains(s.name.as_str()) {
					error!(
						"UI scene `{}` re-uses element name `{}`",
						self.current_scene.as_ref().unwrap(),
						s.name
					);
				} else {
					used_names.insert(s.name.to_string());
				}
				self.elements.push(UiElement::new_sprite(Sprite::new(
					&self.ct,
					s.name.to_string(),
					s.sprite.to_string(),
					s.mask.map(|x| x.to_string()),
					s.rect,
				)));
			} else if v.is::<RadialBuilder>() {
				let r = v.cast::<RadialBuilder>();
				if used_names.contains(r.name.as_str()) {
					error!(
						"UI scene `{}` re-uses element name `{}`",
						self.current_scene.as_ref().unwrap(),
						r.name
					);
				} else {
					used_names.insert(r.name.to_string());
				}
				self.elements.push(UiElement::new_radialbar(RadialBar::new(
					&self.ct,
					r.name.to_string(),
					r.stroke,
					r.color,
					r.rect,
					r.progress,
				)));
			} else if v.is::<TextBoxBuilder>() {
				let t = v.cast::<TextBoxBuilder>();
				if used_names.contains(t.name.as_str()) {
					error!(
						"UI scene `{}` re-uses element name `{}`",
						self.current_scene.as_ref().unwrap(),
						t.name
					);
				} else {
					used_names.insert(t.name.to_string());
				}
				let mut b = TextBox::new(
					state,
					t.name.to_string(),
					t.font_size,
					t.line_height,
					t.font,
					t.justify,
					t.rect,
				);
				b.set_text(state, &t.text);
				self.elements.push(UiElement::new_text(b));
			} else {
				// TODO: better message
				error!("bad type in builder array")
			}
		}
		return Ok(());
	}

	/// Draw all ui elements
	pub fn draw(&mut self, input: Rc<RenderInput>, state: &mut RenderState) -> Result<()> {
		// Initialize start scene if we haven't yet
		if self.current_scene.is_none() {
			self.set_scene(
				state,
				input.clone(),
				self.ct.get_config().start_ui_scene.clone(),
			)?;
		}

		// Update timings if they're being displayed
		if self.show_timings {
			self.fps_indicator.step(&input, state);
		}

		let mut keep_doing_actions = true;

		// Send player state change events
		if {
			let player = input.player.ship;
			if let Some(player) = player {
				let ship = input.phys_img.get_ship(&PhysSimShipHandle(player)).unwrap();
				if self.last_player_state == 0
					|| NonZeroU32::new(self.last_player_state).unwrap()
						!= ship.ship.get_data().get_state().as_int()
				{
					self.last_player_state = ship.ship.get_data().get_state().as_int().into();
					true
				} else {
					false
				}
			} else {
				self.last_player_state = 0;
				true
			}
		} {
			let action: Dynamic = self
				.engine
				.call_fn(
					&mut self.scope,
					&self
						.ct
						.get_config()
						.ui_scenes
						.get(self.current_scene.as_ref().unwrap())
						.unwrap(),
					"event",
					(State::new(input.clone()), PlayerShipStateEvent {}),
				)
				.with_context(|| format!("while handling player state change event"))
				.with_context(|| {
					format!("in ui scene `{}`", self.current_scene.as_ref().unwrap())
				})?;

			if let Some(action) = action.try_cast::<SceneAction>() {
				keep_doing_actions = self.handle_action(state, input.clone(), action)?;
			}
		}

		for i in 0..self.elements.len() {
			match &self.elements[i] {
				UiElement::Sprite(e) => {
					// Draw and update sprites
					let mut sprite = (*e).borrow_mut();
					sprite.step(&input, state);
					sprite.push_to_buffer(&input, state);

					if !keep_doing_actions {
						continue;
					}

					let event = sprite.check_events(&input, state);
					// we MUST drop here, since script calls mutate the sprite RefCell
					drop(sprite);

					match event {
						Event::None => continue,
						_ => {}
					}

					let event_arg = match event {
						Event::None => unreachable!("this shouldn't happen"),

						Event::MouseClick => Dynamic::from(MouseClickEvent {
							down: true,
							element: SpriteElement::new(self.ct.clone(), e.clone()),
						}),

						Event::MouseRelease => Dynamic::from(MouseClickEvent {
							down: false,
							element: SpriteElement::new(self.ct.clone(), e.clone()),
						}),

						Event::MouseHover => Dynamic::from(MouseHoverEvent {
							enter: true,
							element: SpriteElement::new(self.ct.clone(), e.clone()),
						}),

						Event::MouseUnhover => Dynamic::from(MouseHoverEvent {
							enter: false,
							element: SpriteElement::new(self.ct.clone(), e.clone()),
						}),
					};

					let action: Dynamic = self
						.engine
						.call_fn(
							&mut self.scope,
							&self
								.ct
								.get_config()
								.ui_scenes
								.get(self.current_scene.as_ref().unwrap())
								.unwrap(),
							"event",
							(State::new(input.clone()), event_arg),
						)
						.with_context(|| format!("while handling event `{:?}`", event))
						.with_context(|| {
							format!("in ui scene `{}`", self.current_scene.as_ref().unwrap())
						})?;

					if let Some(action) = action.try_cast::<SceneAction>() {
						keep_doing_actions = self.handle_action(state, input.clone(), action)?;
					}
				}

				UiElement::RadialBar(e) => {
					// Draw and update radialbar
					let mut x = (*e).borrow_mut();
					x.step(&input, state);
					x.push_to_buffer(&input, state);
				}

				UiElement::Text(..) => {}
			}
		}

		return Ok(());
	}

	/// Do a SceneAction.
	/// Returns true if we should execute the action that is next, and false otherwise.
	fn handle_action(
		&mut self,
		state: &mut RenderState,
		input: Rc<RenderInput>,
		action: SceneAction,
	) -> Result<bool> {
		Ok(match action {
			SceneAction::None => true,
			SceneAction::GoTo(s) => {
				self.set_scene(state, input.clone(), s.clone())?;
				false
			}
		})
	}
}

impl<'a> UiManager {
	/// Get textareas
	pub fn get_textareas(
		&'a mut self,
		input: &RenderInput,
		state: &RenderState,
	) -> Vec<TextArea<'a>> {
		let mut v = Vec::with_capacity(32);

		if self.current_scene.is_none() {
			return v;
		}

		if self.show_timings {
			v.push(self.fps_indicator.get_textarea(state, input))
		}

		for t in self
			.elements
			.iter()
			.map(|x| x.text())
			.filter(|x| x.is_some())
			.map(|x| x.unwrap())
			.map(|x| x.get_textarea(state, input))
		{
			v.push(t)
		}

		v
	}
}
