use glyphon::{cosmic_text::Align, Attrs, Buffer, Color, Metrics, Shaping, TextArea, TextBounds};
use nalgebra::Vector2;

use super::super::api::{Rect, TextBoxFont, TextBoxJustify};
use crate::{RenderInput, RenderState};

#[derive(Debug)]
pub struct TextBox {
	pub name: String,
	pub font: TextBoxFont,
	pub justify: TextBoxJustify,
	pub rect: Rect,
	pub buffer: Buffer,
}

impl TextBox {
	pub fn new(
		state: &mut RenderState,
		name: String,
		font_size: f32,
		line_height: f32,
		font: TextBoxFont,
		justify: TextBoxJustify,
		rect: Rect,
	) -> Self {
		let mut buffer = Buffer::new(
			&mut state.text_font_system,
			Metrics::new(font_size, line_height),
		);

		// Do NOT apply UI scale here, that's only done when we make a TextArea
		buffer.set_size(&mut state.text_font_system, rect.dim.x, rect.dim.y);

		Self {
			name,
			font,
			justify,
			rect,
			buffer,
		}
	}

	pub fn set_text(&mut self, state: &mut RenderState, text: &str) {
		let mut attrs = Attrs::new();
		attrs = match self.font {
			TextBoxFont::Monospace => attrs.family(glyphon::Family::Monospace),
			TextBoxFont::SansSerif => attrs.family(glyphon::Family::SansSerif),
			TextBoxFont::Serif => attrs.family(glyphon::Family::Serif),
		};

		self.buffer
			.set_text(&mut state.text_font_system, text, attrs, Shaping::Advanced);

		for l in &mut self.buffer.lines {
			l.set_align(Some(match self.justify {
				TextBoxJustify::Center => Align::Center,
				TextBoxJustify::Left => Align::Left,
			}));
		}

		self.buffer.shape_until_scroll(&mut state.text_font_system);
	}
}

impl<'a, 'b: 'a> TextBox {
	pub fn get_textarea(&'b self, state: &RenderState, input: &RenderInput) -> TextArea<'a> {
		let rect = self.rect.to_centered(state, input.ct.get_config().ui_scale);

		// Glypon works with physical pixels, so we must do some conversion
		let fac = state.window.scale_factor() as f32;
		let corner_ne = Vector2::new(
			(rect.pos.x - rect.dim.x / 2.0) * fac + state.window_size.width as f32 / 2.0,
			state.window_size.height as f32 / 2.0 - (rect.pos.y * fac + rect.dim.y / 2.0),
		);
		let corner_sw = corner_ne + rect.dim * fac;

		TextArea {
			buffer: &self.buffer,
			top: corner_ne.y,
			left: corner_ne.x,
			scale: input.ct.get_config().ui_scale,
			bounds: TextBounds {
				top: (corner_ne.y) as i32,
				bottom: (corner_sw.y) as i32,
				left: (corner_ne.x) as i32,
				right: (corner_sw.x) as i32,
			},
			default_color: Color::rgb(255, 255, 255),
		}
	}
}
